/********************************************************
                                         2009 1/1 
                                         2011 3/19 update
   Linux KeyHoleTV(C) 2008,2009          OISEYER Inc.
                                       Takashi Kosaka
   Linux ESD Sound Dynamic Link Library

   File: ESDObj.c

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <time.h>
#include <sys/time.h>
#include <esd.h>

#define   MAX_AUDIO_NUM                 20
#define   AUDIO_44100_100_NUM         10240    /* 1000 mill sec 44100 Hz Bytes  */
#define   AUDIO_NOTIFY_SIZE_2          4410    /* 16 bits 44100 Hz 100 Mill Sec short int */

/*****************************************************************
  Sound Object
*****************************************************************/
typedef struct _SoundObj {
	void        *Sample;
	void        *Handle;

	unsigned char *(*GetAudioBuffer)(void *OO);
	void           (*SetNextWriteBuffer)(void *OO,int Size);
	void           (*SoundFunction)(void *OO);
	void           (*StopAuidoFunction)(void *OO);
	void           (*FinishAuidoFunction)(void *OO);

} DataSoundObj, *PSoundObj;

/**********************************************************
 Audio Buffer
***********************************************************/
typedef struct _AudioBuffer {
  unsigned char *Buf;   
  int            Size;
  int            Written;
  struct _AudioBuffer *Next;
} AudioBuffer, *PAudioBuffer;

/************************************************************
  ESD Object
************************************************************/
typedef struct _ESDObj {
  int          Esd;      // ESD Handle
  int          StopIDW;
  int          StopIDR;

  unsigned int Store;
  unsigned int Used;
  int          Start;
	
  PAudioBuffer Read;
  PAudioBuffer Write;

	char        *JoinRoom;
	char        *Exiting;

} ESDObj, *PESDObj;

/*************************************************************
   Set ESD Driver
*************************************************************/
static int SetESD(PESDObj Obj)
{
  esd_format_t   Format;
  int fl;

  Format = ESD_BITS16 | ESD_MONO | ESD_STREAM | ESD_RECORD;

  Obj->Esd = -1;
  if((Obj->Esd =  esd_play_stream_fallback(Format,44100, 0, "LKeyHole")) < 0 ) {
		fprintf(stderr,"esd_play_stream_fallback ERROR\n");
    return(-1);
  }
  
  if ((fl = fcntl(Obj->Esd, F_GETFL)) >= 0) {
    fcntl(Obj->Esd, F_SETFL, O_NDELAY|fl);
  }

  return(0);
}

/********************************************************
  Store Last Sound Data
********************************************************/
static int LastSoundData = 0;

/**********************************************************
  Write NULL Sound
**********************************************************/
static void WriteNullSound(short *Buf,int aaa)
{
  int i;
  short int DD;
  double Sa,LastD;

  if( LastSoundData == 0 ) {
    for( i = 0 ; i < aaa ; i++ ) {
      *(Buf + i) = 0;
    }
    return;
  }

  LastD = (double)LastSoundData;
  Sa = (double)abs(LastSoundData) / (double)aaa;

  //fprintf(stderr,"Last %d Sa:%e\n",LastSoundData,Sa);

  if( LastSoundData > 0 ) {
    LastD = -LastD + Sa;
  }
  else {
    LastD = -LastD - Sa;
  }
  
  for( i = 0 ; i < aaa ; i++ ) {
    DD = (short)LastD;
    *(Buf + i) = DD;

    if( LastD > 0.0 ) {
      LastD = -LastD + Sa;
    }
    else if (LastD < 0.0 ) {
      LastD = -LastD - Sa;
    }
  }
  LastSoundData = 0;
}


/************************************************************
   Write Sound to ESD Driver
************************************************************/
static void WriteSoundToEsdDriver(PESDObj Obj)
{
  fd_set            WriteFd;
  int               Ret,W,bb;
  struct            timeval TT;

  if( Obj->Esd > 0 ) {
    short Buf[AUDIO_NOTIFY_SIZE_2];

    if( Obj->Read != Obj->Write ) {
      FD_ZERO(&WriteFd);
      TT.tv_sec  = 0;
      TT.tv_usec = 0;
      W = Obj->Esd + 1;
      FD_SET(Obj->Esd,&WriteFd);

      if((Ret = select(W,NULL,&WriteFd,NULL,&TT)) > 0 ) {

	if( FD_ISSET(Obj->Esd,&WriteFd) ) { /* $B=q$-9~$_5v2D(B */
	  
	  bb = write(Obj->Esd,Obj->Read->Buf + Obj->Read->Written,
		     Obj->Read->Size - Obj->Read->Written);
	  
	  //fprintf(stderr,"Write: %d\n",bb);
	  if( bb > 0 ) {
	    Obj->Read->Written += bb;
	    if( Obj->Read->Written >= Obj->Read->Size ) {
	      Obj->Used++;
	      Obj->Read->Written = Obj->Read->Size = 0;
	      Obj->Read = Obj->Read->Next;
	    }
	  }
	}
      }
    }
    else {

      FD_ZERO(&WriteFd);
      TT.tv_sec  = 0;
      TT.tv_usec = 0;
      W = Obj->Esd + 1;
      FD_SET(Obj->Esd,&WriteFd);

      if((Ret = select(W,NULL,&WriteFd,NULL,&TT)) > 0 ) {

	if( FD_ISSET(Obj->Esd,&WriteFd) ) { /* $B=q$-9~$_5v2D(B */
	  WriteNullSound(Buf,AUDIO_NOTIFY_SIZE_2);
	  write(Obj->Esd,Buf,AUDIO_NOTIFY_SIZE_2 * 2);
	}
      }
    }
  }
}
  
/************************************************************
  Interface Functions
************************************************************/

/************************************************************
  Initialize ESD Driver And Allocate Memory
  Call by Main Thread
************************************************************/
static void * InitEsdDriver(char *Exiting, char *JoinRoom)
{
  PESDObj      Obj;
  int          i;
  PAudioBuffer Now,Prv,Top;		
	
  if((Obj = (PESDObj)malloc(sizeof(ESDObj))) == NULL ) {
		fprintf(stderr,"malloc ERROR\n");
    return((void *)NULL);
  }

  if( SetESD(Obj) < 0 ) {
    return((void *)NULL);
  }

  Obj->StopIDW = Obj->StopIDR = 0;
  Obj->Store = Obj->Used = Obj->Start = 0;
	Obj->JoinRoom = JoinRoom;
	Obj->Exiting  = Exiting;

  Top = Prv = NULL;

  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    if((Now = (PAudioBuffer)malloc(sizeof(AudioBuffer))) == NULL ) {
      return((void *)NULL);
    }

    if((Now->Buf = (unsigned char *)malloc(AUDIO_44100_100_NUM)) == NULL ) {
      return((void *)NULL);
    }
    Now->Size = Now->Written = 0;

    if( ! Top ) {
      Top = Now;
    }

    if( Prv ) {
      Prv->Next = Now;
    }

    Prv = Now;
  }
  Now->Next = Top;

  /* Store Ring Buffer */
  Obj->Read = Obj->Write = Now;

  return((void *)Obj);
}

/***********************************************************
   Get Wirte Buffer
   Call by Thread 1
***********************************************************/
static unsigned char *GetWriteESDBuffer(void *OO)
{
  PESDObj            Obj;
  Obj = (PESDObj)OO;
  return((unsigned char *)Obj->Write->Buf);
}

/**********************************************************
  Set Next Write Buffer
   Call by Thread 1
**********************************************************/
static void SetNextESDWriteBuffer(void *OO,int Size)
{
  PESDObj            Obj;
  Obj = (PESDObj)OO;
  Obj->Write->Size = Size;
  Obj->Write->Written = 0;
  Obj->Write = Obj->Write->Next;
  Obj->Store++;
}

/**********************************************************
  Sound Function
   Call by Thread 2
**********************************************************/
static void SoundESDFunction(void *OO)
{
  PESDObj            Obj;
  Obj = (PESDObj)OO;

  if( Obj->StopIDW != Obj->StopIDR ) {

    /* $BDd;_MW5a$,Mh$?(B */
    if( Obj->Esd > 0 ) {
      esd_close(Obj->Esd);
      Obj->Esd = -1;
    }
		
    Obj->Read = Obj->Write;
    Obj->Read->Size = Obj->Read->Written = 0;
    Obj->Start   = 0;
    Obj->StopIDR = Obj->StopIDW;
    Obj->Used    = Obj->Store = 0;
  }

  if( *Obj->Exiting == 0 && *Obj->JoinRoom ) {
		
    if( Obj->Esd < 0 ) {
      SetESD(Obj);
    }

    if(  Obj->Start == 0 ) {	
      if( (Obj->Used + 5) < Obj->Store  ) {	
	Obj->Start = 1;
	WriteSoundToEsdDriver(Obj);
      }
    }
    else {
      WriteSoundToEsdDriver(Obj);
    }
  }
}

/********************************************************
  Stop ESD Audio
  Called by Main Thread
********************************************************/
static void StopESDAudio(void *OO)
{
  PESDObj            Obj;
  Obj = (PESDObj)OO;
  Obj->StopIDW++;
}

/********************************************************
  Finish ESD Audio
  Called by Main Thread
********************************************************/
static void FinishESDAudio(void *OO)
{
  PESDObj             Obj;
  PAudioBuffer        Now,Next;
  int                 i;
  Obj = (PESDObj)OO;

  if( Obj->Esd > 0 ) {
    esd_close(Obj->Esd);
  }
  Now = Obj->Read;
  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    Next = Now->Next;
    free((void *)Now);
    Now = Next;
  }
  free((void *)Obj);
}

/***************************************************************
 Set ESD Driver

 This function called by dlsym(Handle,"SetAudioDriver");
 So, propgramers can not change function mame and arguments.
 Called by Main Thread.
***************************************************************/
int SetAudioDriver(PSoundObj Obj,char *exiting,char *Join)
{
  if((Obj->Handle = InitEsdDriver(exiting,Join)) == NULL ) {
    return(1);
  }
  Obj->GetAudioBuffer     = GetWriteESDBuffer;
  Obj->SetNextWriteBuffer = SetNextESDWriteBuffer;
  Obj->SoundFunction      = SoundESDFunction;
  Obj->StopAuidoFunction  = StopESDAudio;
  Obj->FinishAuidoFunction = FinishESDAudio;
  return(0);
}
